import os
from os import R_OK, W_OK, X_OK, fork, symlink, unlink, O_CREAT, O_EXCL, readlink, chown, chmod
from os import setgroups, setgid, setuid, chdir, umask, execvpe, execv, waitpid, WEXITSTATUS
from os import getpid, getuid, _exit
from os.path import join, basename, realpath, lexists, islink
from subprocess import check_call
from stat import S_IFDIR
from tempfile import mkdtemp
import resource

from rainbow.util import Checker, mount, make_dirs, get_fds, read_envdir

USAGE = """

NAME

  rainbow-inject

SYNOPSIS

  rainbow-inject is a security-wrapper for Sugar Activities. Its purpose is to
  launch them in a low-privilege environment so that, if they are exploited,
  they are able to do less harm than would be the case if they ran with the
  full authority of the XO's human operator.

  This isolation is accomplished by running each activity as a separate user
  from a set of users that is known, at system installation time, to have
  tightly restricted access to the file-system.

  The environment that rainbow-inject creates for each activity it starts is
  supposed to be controlled by an instance-specific security-record (usually
  derived from a prototype record stored per-bundle) but this is not yet
  implemented. Therefore, Rainbow currently offers no way to customize what
  kinds of isolation it provides.

ENVIRONMENT VARIABLES

  SUGAR_BUNDLE_PATH    : MUST contain an absolute path to a bundle to launch.

  SUGAR_BUNDLE_ID      : MUST contain the unique bundle-id of the bundle at
                         SUGAR_BUNDLE_PATH.

  RAINBOW_DIR          : Optional; defaults to $HOME/isolation/1
                         This directory is used by rainbow-inject for all its
                         run-time data: currently, this means uid/gid
                         allocation data and for storing persistent
                         activity-data storage

  RAINBOW_STRACE_LOG   : Optional; defaults to Nothing.
                         If set, rainbow-inject will cause strace to write a
                         complete log to stdout.

  RAINBOW_CONSTANT_UID : Optional; defaults to Nothing.
                         If set, rainbow-inject will cause strace to write a
                         complete log to stdout.

ARGUMENTS

  Any arguments will be exec'ed verbatim after rainbow-inject properly
  configures its environment.

FILE DESCRIPTORS

  rainbow-inject will close all open file descriptors except those marked
  'safe' by its caller.

  rainbow-inject will print trace information on descriptor 1 and may print
  error messages on descriptor 2.

RETURN CODES

  Not specified.

"""

# Locations of these two files come from the XAUTHORITY and ICEAUTHORITY
# environment variables which are, alas, not set yet when rainbow starts.
# So they're hard-coded here; values should match those set in
# olpc-dm (in olpc-utils package)
_AUTH_HOME='/var/tmp/olpc-auth'
_XAUTHORITY=_AUTH_HOME + '/.Xauthority'
_ICEAUTHORITY=_AUTH_HOME + '/.ICEauthority'

def strace(log, argv, env):
    log(1, 'applying strace')
    args = ['/usr/bin/strace', '-f']
    for k, v in env.iteritems():
        args.append('-E')
        args.append('%s=%s' % (k, v))
    args.extend(argv)

    return args, {}

def reserve_elt(pool_dir, elt, max_elt, incr, elt_name):
    fd = None
    while elt < max_elt:
        try:
            fd = os.open(join(pool_dir, str(elt)), O_EXCL | O_CREAT, 0600)
            os.close(fd)
            break
        except OSError:
            elt += incr
    if fd is None:
        raise RuntimeError("No " + elt_name + " available.")
    return elt

def reserve_credentials(log, spool, data_id, pset):
    # Pick a gid for yourself.
    gid = reserve_elt(join(spool, 'gid_pool'), 10001, 65534, 2, 'gids')

    # Then try to atomically symlink the gid you picked to 'name' in gid_dir.
    # If you succeed, then you have the right gid.
    # If you fail, someone else has the right gid so release yours and use theirs.
    name_path = join(spool, 'bundle_id_to_gid', data_id)
    gid_path = join(spool, 'gid_pool', str(gid))
    try:
        symlink(gid_path, name_path)
    except OSError:
        unlink(gid_path)
        gid = int(basename(realpath(name_path)))

    # Decide whether to reserve a new (even) uid or to set uid <- gid and reserve that.
    uid = pset.has_permission('constant-uid') and gid or reserve_elt(join(spool, 'uid_pool'), 10000, 65534, 2, 'uids')
    open(join(spool, 'uid_pool', str(gid)), 'w').close()
    log(1, 'reserved credentials (%d, %d)', uid, gid)

    path = join(spool, 'uid_to_gid', str(uid))
    if not lexists(path):
        symlink(str(gid), path)
    else:
        assert islink(path) and readlink(path) == str(gid)

    return (uid, gid)

def grab_home(log, spool, data_id, uid, gid, owner_gid):
    home = join(spool, 'uid_to_home_dir', str(uid))
    make_dirs(home, uid, owner_gid, 0770)
    chown(home, uid, owner_gid)
    # Per discussion with Bert Freudenberg, set the setgid bit on $SAR/instance
    # (i.e. $HOME) so that Sugar can better write inside it. <MS>
    chmod(home, 02770)
    return home

def configure_home(spool, owner_uid, owner_gid, uid, gid, home):
    # Beware of CONSTANT_UID + instance==home: the instance dir may already exist.
    path = join(spool, 'uid_to_instance_dir', str(uid))
    if not lexists(path):
        symlink(home, path)

    if not lexists(join(home, 'instance')):
        symlink('.', join(home, 'instance'))

    path = join(spool, 'gid_to_data_dir', str(gid))
    make_dirs(path, owner_uid, gid, 0770)
    chown(path, owner_uid, gid)
    chmod(path, 0770)
    if not lexists(join(home, 'data')):
        symlink(path, join(home, 'data'))

    #XXX: ARGH pippy. <MS>
    make_dirs(join(home, 'tmp'), uid, gid, 0700)
    mount('tmpfs', join(home, 'tmp'), 'tmpfs', 0, '')

def mount_fsen(log, home):
    log(1, 'Mounting tmpfsen on /tmp and /var/tmp.')
    #XXX: Kills X socket. :( <MS>
    #mount('tmpfs', '/tmp', 'tmpfs', 0, '')
    mount('tmpfs', '/var/tmp', 'tmpfs', 0, '')

def run_assistant(log, assistant, env, owner_uid, owner_gid, uid, groups):
    if assistant:
        envdir = None
        try:
            envdir = mkdtemp()
            chown(envdir, owner_uid, owner_gid)
            pid = fork()
        except:
            if envdir: check_call(['/bin/rm', '-rf', envdir])
            raise
        else:
            if not pid:
                log(1, 'Dropping privilege to run assistant.')
                setgroups(groups)
                setgid(owner_gid)
                setuid(owner_uid)
                log(1, 'Running assistant.')
                assistant_argv = [assistant, '-v', '-v', '-v', '-u', str(uid), '-e', envdir]
                log(1, '%r %r', assistant_argv, env)
                execvpe(assistant_argv[0], assistant_argv, env)
                _exit(55)
            else:
                try:
                    pid, status = waitpid(pid, 0)
                    log(1, 'Assistant returned %d.', status)
                    assert not WEXITSTATUS(status)
                    return read_envdir(envdir)
                finally:
                    log(1, 'pid %d  uid %d', getpid(), getuid())
                    if envdir: check_call(['/bin/rm', '-rf', envdir])

def launch(log, home, uid, gid, groups, argv, env, cwd, pset, safe_fds):
    # Set appropriate group membership(s), depending on requested permissions
    log(1, 'dropping privilege to (%d, %d, %r)', uid, gid, groups)
    setgroups(groups)
    setgid(gid)
    setuid(uid)

    # Limit various resources
    # Must be done *after* setting uid/gid
    # This should come from the permissions.info file, but this is OK
    #   for now.
    try:
        def set_limit(lim_name, unix_name):
            p = pset.permission_params('lim_' + lim_name)
            if p != None:
                x = int(float(p[0]))
                y = int(float(x*1.10))
                log(1, 'Setting RLIMIT_%s to %d,%d', unix_name, x, y)
                resource.setrlimit(getattr(resource, 'RLIMIT_'+unix_name), (x,y))

        set_limit('nofile', 'NOFILE')
        set_limit('fsize', 'FSIZE')
        set_limit('mem', 'AS')
        set_limit('nproc', 'NPROC')

    except:
        pass # Why would we be throwing exceptions when setting rlimits? <MS>

    log(1, 'chdir to %s' % cwd)
    chdir(cwd)

    log(1, 'umask(0)')
    umask(0)

    if pset.has_permission('strace'):
        argv, env = strace(log, argv, env)

    log(1, 'about to execve\nargv: %s\nenv: %s', argv, env)
    log(1, 'closing all fds but %s', safe_fds)
    for fd in get_fds():
        if fd not in safe_fds:
            try: os.close(fd) # propagate failure from EIO or EBADF.
            except: pass

    execvpe(argv[0], argv, env)

def check_data_id(data_id):
    # XXX: How do I figure out what MAX_PATH_LEN is in python? <MS>
    # XXX: How long are GECOS fields permitted to be? <MS>
    assert data_id and '\0' not in data_id and len(data_id) < 128

def check_argv(argv):
    assert argv

def check_cwd(uid, gid, cwd):
    ck = Checker(cwd, uid, gid)
    assert ck.positive(R_OK | X_OK, S_IFDIR)

def check_spool(spool, owner_uid, owner_gid):
    make_dirs(spool, 0, 0, 0755)
    spool_dirs = ('uid_pool', 'gid_pool', 'uid_to_gid', 'bundle_id_to_gid',
                  'uid_to_instance_dir', 'gid_to_data_dir', 'uid_to_home_dir')
    for frag in spool_dirs:
        make_dirs(join(spool, frag), 0, 0, 0755)
        ck = Checker(join(spool, frag), owner_uid, owner_gid)
        assert ck.positive(R_OK | X_OK, S_IFDIR)

def check_owner(owner_uid, owner_gid):
    return True

def check_home_dirs(uid, gid, home):
    for frag in ('instance', 'data', 'tmp'):
        ck = Checker(join(home, frag), uid, gid)
        assert ck.positive(R_OK | W_OK | X_OK, S_IFDIR)

def check_home(uid, gid, home):
    ck = Checker(home, uid, gid)
    assert ck.positive(R_OK | W_OK | X_OK, S_IFDIR)

def configure_groups(groups, gid):
    groups.insert(0, gid)
    return list(set(groups))

def inject(log, spool, env, argv, cwd, pset, safe_fds, owner_uid, owner_gid,
           groups, data_id, assistant):
    # Note: exceptions are intended to bubble up to the caller and should
    # terminate execution.
    check_data_id(data_id)
    check_argv(argv)
    check_owner(owner_uid, owner_gid)

    check_spool(spool, owner_uid, owner_gid)

    uid, gid = reserve_credentials(log, spool, data_id, pset)
    home = grab_home(log, spool, data_id, uid, gid, owner_gid)
    configure_home(spool, owner_uid, owner_gid, uid, gid, home)

    if cwd is None:
        cwd = home
    check_cwd(uid, gid, cwd)
    check_home_dirs(uid, gid, home)
    check_home_dirs(owner_uid, owner_gid, home)
    check_home(uid, gid, home)

    groups = configure_groups(groups, gid)
    env_updates = run_assistant(log, assistant, env, owner_uid, owner_gid, uid, groups)
    if env_updates: env.update(env_updates)

    mount_fsen(log, home)

    launch(log, home, uid, gid, groups, argv, env, cwd, pset, safe_fds)
